# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.plugins.task.runners.rps import *  # noqa: F401,F403
from rally.plugins.task.runners import rps as _new

# import it as last item to be sure that we use the right module
from rally.common import logging


logging.log_deprecated_module(
    target=__name__, new_module=_new.__name__, release="3.0.0"
)
